/*
 * MiniXML.java
 *
 * Created on July 30, 2001, 7:48 AM
 */

package com.hbaastrup.MiniXML;

/**
 *
 * @author  Henrik Baastrup (c) 2001
 * @version 0.1
 *
 * Home page: http://digilander.iol.it/hba/Linux/index.html
 * This software is licensed under the GNU LGPL (Lesser general Public License):
 *   http://www.gnu.org/copyleft/lesser.html
 *
 */
import java.io.*;
import java.util.*;


public class MiniXML 
{
    public static final String VERSION = "0.8";
    public static final int NEW_TAG_STATE = 0; 
    public static final int GET_TAG_NAME_STATE = 1;
    public static final int GET_ATTR_NAME_STATE = 2;
    public static final int GET_ATTR_VALUE_STATE = 3;
    public static final int GET_TAG_VALUE_STATE = 4;
    public static final int NEW_SUB_TAG_STATE = 5;
    public static final int GET_END_TAG_NAME_STATE = 6;
    public static final int TAG_END_STATE = 7;
    public static final int TAG_COMMENT_STATE = 8;
    public static final int TAG_DECLARATION_STATE = 9;

    Element xmlTree;
    
    public MiniXML() {
        init();
    }//miniXML
    
    public void init()
    {
        xmlTree = new Element();
    }//init
    
    private String parser(String xmlStr, Element elm)
    {
        byte bStr[];
        byte tagName[] = new byte[1];
        byte attrName[] = new byte[1];
        byte attrValue[] = new byte[1];
        byte tagValue[] = new byte[1];
        byte endTagName[] = new byte[1];
        String subParse;
        String parseStr;
        Element subElm;
        ElementAttribute attr = new ElementAttribute();
        int state = NEW_TAG_STATE;
        int oldState, pingState, i,j;
        boolean ignore, stop, noStep;
        
        parseStr = xmlStr;
        bStr = parseStr.getBytes();
        pingState = 0;
        oldState = 0;
        stop = false;
        ignore = false;
        noStep = false;
        i=0;
        j=0;
        while (i<bStr.length && !stop) {
            switch (state) {
                case NEW_TAG_STATE: //new tag
                    if (ignore) { //If we are in an ignore state change state for the next
                        if (bStr[i]=='>') ignore = false;
                    }
                    else if (bStr[i]=='<') {
                        if (bStr[i+1]=='/') {
                            endTagName = new byte[bStr.length-i];
                            endTagName[0] = '\0';
                            j = 0;
                            i++;
                            state = GET_END_TAG_NAME_STATE;
                        }
                        else if (bStr[i+1]=='?') {
                            oldState = state;
                            state = TAG_DECLARATION_STATE;
                        }
                        else if (bStr[i+1]=='!') {
                            if (bStr[i+2]=='-' && bStr[i+3]=='-') {
                                oldState = state;
                                state = TAG_COMMENT_STATE;
                            }
                            else ignore = true;
                        }
                        else {
                            tagName = new byte[bStr.length-i];
                            tagName[0] = '\0';
                            j = 0;
                            state = GET_TAG_NAME_STATE;
                        }
                    }
                    break;
                    
                case GET_TAG_NAME_STATE://get begin tag name
                    if (bStr[i]==' ' || bStr[i]=='\t') {
                        tagName[j] = '\0';
                        elm.setName(tagName,j);
                        j=0;
                        attrName = new byte[bStr.length-i];
                        attrName[0] = '\0';
                        state = GET_ATTR_NAME_STATE;
                        break;
                    }
                    if (bStr[i]=='>') {
                        tagName[j] = '\0';
                        elm.setName(tagName,j);
                        j=0;
                        tagValue = new byte[bStr.length-i];
                        tagValue[0] = '\0';
                        state = GET_TAG_VALUE_STATE;
                        break;
                    }
                    if (bStr[i]=='/' && bStr[i+1]=='>') {
                        tagName[j] = '\0';
                        elm.setName(tagName,j);
                        i++;
                        state = TAG_END_STATE;
                        break;
                    }
                    tagName[j] = bStr[i];
                    j++;
                    break;
                    
                case GET_ATTR_NAME_STATE: //Get attribute name
                    if (bStr[i]==' ' || bStr[i]=='\t') break;
                    if (bStr[i]=='=') {
                        attrName[j] = '\0';
                        attr.name = new String(attrName,0,j);
                        j=0;
                        attrValue = new byte[bStr.length-i];
                        attrValue[0] = '\0';
                        state = GET_ATTR_VALUE_STATE;
                        break;
                    }
                    if (bStr[i]=='>') {
                        attrName[j] = '\0';
                        attr.name = new String(attrName,0,j);
                        elm.addAttribute(attr);
                        attr = new ElementAttribute();
                        tagValue = new byte[bStr.length-i];
                        tagValue[0] = '\0';
                        state = GET_TAG_VALUE_STATE;
                        break;
                    }
                    if (bStr[i]=='/' && bStr[i+1]=='>') {
                        attrName[j] = '\0';
                        attr.name = new String(attrName,0,j);
                        elm.addAttribute(attr);
                        attr = new ElementAttribute();
                        i++;
                        state = TAG_END_STATE;
                        break;
                    }
                    attrName[j] = bStr[i];
                    j++;
                    break;
                    
                case GET_ATTR_VALUE_STATE: //Get attribute value
                    if (bStr[i]=='\'' || bStr[i]=='\"') {
                        boolean cont = false;
                        switch (pingState) {
                            case 0:
                                if (bStr[i]=='\'') pingState = 1; // we have a ' and expect the attribute end with a '
                                else if (bStr[i]=='\"') pingState = 2; // we have a " and expect the attribute end with a "
                                break;
                            case 1:
                                if (bStr[i]=='\'') pingState = 0;
                                else cont = true;
                                break;
                            case 2:
                                if (bStr[i]=='\"') pingState = 0;
                                else cont = true;
                                break;
                        }
                        if (!cont) break;
                    }
                    if ((bStr[i]==' ' || bStr[i]=='>' || bStr[i]=='/') && pingState==0) { // read the next cahr of the attribute value
                        attrValue[j] = '\0';
                        attr.value = new String(attrValue,0,j);
                        elm.addAttribute(attr);
                        attr = new ElementAttribute();
                        if (bStr[i]==' ') {
                            j=0;
                            attrName = new byte[bStr.length-i];
                            attrName[0] = '\0';
                            state = GET_ATTR_NAME_STATE;
                            break;
                        }
                        if (bStr[i]=='>') {
                            j=0;
                            tagValue = new byte[bStr.length-i];
                            tagValue[0] = '\0';
                            state = GET_TAG_VALUE_STATE;
                            break;
                        }
                        if (bStr[i]=='/' && bStr[i+1]=='>') {
                            i++;
                            state = TAG_END_STATE;
                            break;
                        }
                    }
                    attrValue[j] = bStr[i];
                    j++;
                    break;
                    
                case GET_TAG_VALUE_STATE: //Get tag value
                    if (bStr[i]=='<') {
                        if (bStr[i+1]=='!' && bStr[i+2]=='-' && bStr[i+3]=='-') {// This is a comment!
                            oldState = state;
                            state = TAG_COMMENT_STATE;
                            break;
                        }
                        tagValue[j] = '\0';
                        elm.setValue(tagValue,j);
                        if (bStr[i+1]=='/') {
                            i++;
                            endTagName = new byte[bStr.length-i];
                            endTagName[0] = '\0';
                            j = 0;
                            state = GET_END_TAG_NAME_STATE;
                            break;
                        }
                        i--;
                        state = NEW_SUB_TAG_STATE;
                        break;
                    }
                    tagValue[j] = bStr[i];
                    j++;
                    break;
                    
                case NEW_SUB_TAG_STATE: //Start to get sub tags from the actual tag
                    if (bStr[i]=='<') {
                        if (bStr[i+1]=='!' && bStr[i+2]=='-' && bStr[i+3]=='-') {
                            oldState = state;
                            state = TAG_COMMENT_STATE;
                            break;
                        }
                        if (bStr[i+1]=='/') {
                            endTagName = new byte[bStr.length-i];
                            endTagName[0] = '\0';
                            j = 0;
                            i++;
                            state = GET_END_TAG_NAME_STATE;
                            break;
                        }
                        subParse = parseStr.substring(i);
                        subElm = new Element();
                        elm.addElement(subElm);
                        if ((parseStr=parser(subParse, subElm))==null) {
                            bStr[i] = '\0';
                            stop = true;
                            break;
                        }
                        bStr = parseStr.getBytes();
                        i = 0;
                        if (bStr[0]=='<') {
                            if (bStr[1]=='!' && bStr[2]=='-' && bStr[3]=='-') {
                                oldState = state;
                                state = TAG_COMMENT_STATE;
                                break;
                            }
                            noStep = true;
                        }
                    }
                    break;
                    
                case GET_END_TAG_NAME_STATE: //Finish the actual tag
                    if (bStr[i]=='>') {
                        state = TAG_END_STATE;
                        break;
                    }
                    endTagName[j] = bStr[i];
                    j++;
                    break;
                    
                case TAG_END_STATE: //No more values, attrbute or subtag to read. Return!
                    stop = true;
                    break;
                    
                case TAG_COMMENT_STATE: //We read a comment but we do not save it!
                    if (bStr[i]=='-' && bStr[i+1]=='-' && bStr[i+2]=='>') {
                        state = oldState;
                        i += 2;
                    }
                    break;
                    
                case TAG_DECLARATION_STATE: //We read a declaraton but do not save it!
                    if (bStr[i]=='?' && bStr[i+1]=='>') { //end of decleration
                        state = oldState;
                        i++;
                    }
                    break;
            }
            if (noStep) noStep = false;
            else if (bStr[i]!='\0') i++;
            else {
                stop = true;
                break;
            }
        }
        if (stop) {
            if (bStr[i]=='\0') return null;
            i--;
            return parseStr.substring(i);
        }
        
        return null;
    }//parser
    
    private String cleanUpString(String str)
    {
        byte bStr[], retbStr[];
        String retStr;
        int j;
        
        bStr = str.getBytes();
        retbStr = new byte[bStr.length];
        
        j = 0;
        for (int i=0; i< bStr.length; i++) {
            switch (bStr[i]) {
                case '\r':
                case '\n':
                    break;
                    
                default:
                    retbStr[j] = bStr[i];
                    j++;
                    break;
            }
        }
        retStr = new String(retbStr);
        return retStr.trim();
    }//scleanUpString
    
    public Element getTree() { return xmlTree; }
    
    public void parse(String xmlStr)
    {
        String parseStr;
        
        
        parseStr = cleanUpString(xmlStr);
        init();
        parser(parseStr, xmlTree);
    }//parse
    
    
    int iSpace = 0;
    public void writeOutTree(PrintStream out, Element elm)
    {
        Element subElm;
        byte bSpace[];
        String sSpace;
        Vector vect;
        Enumeration attrList;
        ElementAttribute attr;
        
        iSpace++;
        if (elm==null) {
            iSpace--;
            return;
        }
        bSpace = new byte[iSpace];
        for (int i=0; i< iSpace; i++) bSpace[i] = ' ';
        sSpace = new String(bSpace);
        out.println(sSpace+elm.getName()+": "+elm.getValue());
        vect = elm.getAttributes();
        if (vect!=null) {
            attrList = vect.elements();
            while (attrList.hasMoreElements()) {
                attr = (ElementAttribute)attrList.nextElement();
                out.println(sSpace+" attribute:"+attr.name+"="+attr.value);
            }
        }
        subElm = elm.getFirstSubElement();
        while (subElm!=null) {
            writeOutTree(out,subElm);
            subElm = elm.getNextSubElement();
        }
        iSpace--;
    }//writeOutTree
    
    public static void main(String[] args)
    {
        MiniXML miniXML = new MiniXML();
        RandomAccessFile f;
        String s, xmlStr;
        Element tree, subTree;
        
        if (args.length==0) {
            System.out.println("Missing filename");
            return;
        }
        
        xmlStr = "";
        try {
            f = new RandomAccessFile(args[0],"r");
            s ="";
            while (s!=null) {
                s = f.readLine();
                if (s!=null) {
                    System.out.println(s);
                    xmlStr = xmlStr.concat(s);
                }
            }
            f.close();
        }
        catch (IOException ex) {
            System.out.println(ex);
        }
        miniXML.parse(xmlStr);
        miniXML.writeOutTree(System.out, miniXML.getTree());
        System.out.println();
        tree = miniXML.getTree();
        tree.initSearch();
        do {
            subTree = tree.searchFor("pool");
            if (subTree==null) System.out.println("NULL tree");
            else miniXML.writeOutTree(System.out, subTree);
        } while (subTree!=null);
    }//main
}//MiniXML
